// $Id: Client.cc,v 1.20.2.2 1998/11/08 04:56:26 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "ConnMgr.h"
#include "ProgressMeter.h"
#include "Pause.h"
#include "AddrGen.h"
#include "ClXaction.h"
#include "LocalStats.h"
#include "Log.h"
#include "Client.h"

/* ClientCfg */

ClientCfg::ClientCfg(): 
	theConnMgr(0), theProgressMeter(0), theLoader(0), theAddrGen(0) {
}

void ClientCfg::destroy() {
	delete theConnMgr; theConnMgr = 0;
	delete theProgressMeter; theProgressMeter = 0;
	delete theLoader; theLoader = 0;
	delete theAddrGen; theAddrGen = 0;
	theXactCfg.destroy();
}

ClientCfg ClientCfg::clone() const {
	ClientCfg cfg;

	cfg.theConnMgr = theConnMgr->clone();
	cfg.theProgressMeter = theProgressMeter->clone();
	cfg.theLoader = theLoader->clone();
	cfg.theAddrGen = theAddrGen->clone();
	cfg.theXactCfg = theXactCfg.clone();

	return cfg;
}

/* Client */

Client::~Client() {
	theCfg.destroy();
}

Thread *Client::clone() const {
	Client *c = new Client();
	c->configure(theCfg.clone());
	return c;
}

void Client::configure(const ClientCfg &aCfg) {
	theCfg = aCfg;
	theCfg.theLoader->loadee(this);
	theCfg.theConnMgr->owner(this);
}

void Client::init() {
	Journal << "client " << this << " started." << ende;
	theCfg.theLoader->start();
}

void Client::removed() {
	delete theCfg.theLoader; theCfg.theLoader = 0;
}

void Client::beIdle() {
	if (theCfg.theProgressMeter->enough()) {
		Journal << "client " << this << " stopped." << ende;
		beActive(thDone);
	}
}

void Client::taskStarted(Thread *th) {
	TopProcess::taskStarted(th);
	theCfg.theProgressMeter->noteTaskStart();
	if (theCfg.theLoader)
		theCfg.theLoader->loadChange(LoadInfo(theCfg.theProgressMeter->ccLevel(), +1)); // XXX: direct?
}

void Client::taskStopped(Thread *th) {
	TopProcess::taskStopped(th);
	theCfg.theProgressMeter->noteTaskFinish();
	TheWorld.noteCompletion();
	if (theCfg.theLoader)
		theCfg.theLoader->loadChange(LoadInfo(theCfg.theProgressMeter->ccLevel(), -1)); // XXX: direct?
}

void Client::idleRes(ResMgr *mgr) {
	connMgr()->put(mgr); // ?
}

void Client::voidRes(ResMgr *mgr) {
	connMgr()->del(mgr);
	theErr.tick();
	StatInc(theXactErrCount);
}

void Client::changeLoad(const LoadInfo &li) {
	// XXX: perhaps the loader interface should submit a bunch of addresses
	//      and not just the number?
	if (li.valid()) {
		const int delta = li.curDelta(load());
		theLaunchDebt += delta;
		if (theLaunchDebt < 0)
			theLaunchDebt = 0;
		Notepad << "client " << this << " delta: " << delta << " new launch debt: " << theLaunchDebt << ende;
		if (theLaunchDebt)
			beActive(thRunnable);
	}
}

void Client::checkProgress() {
	if (theCfg.theProgressMeter->enough()) {
		delete theCfg.theLoader; theCfg.theLoader = 0;
		if (idle())
			state(thDone);
	} else
	if (theLaunchDebt)
		state(thRunnable);
}

void Client::run() {
	Address addr;
	if (theCfg.theAddrGen->gen(addr)) {
		Notepad << "client " << this << " enough: " << theCfg.theProgressMeter->enough() << ende;
		while (theLaunchDebt > 0 && !theCfg.theProgressMeter->enough()) {
			theLaunchDebt--;
			if (!launchXact(addr)) {
				theCfg.theProgressMeter->noteTaskFailed();
				TheWorld.noteCompletion(); // count failures too!
				if (theCfg.theLoader)
					theCfg.theLoader->loadChange(LoadInfo(theCfg.theProgressMeter->ccLevel(), 0));
				theErr.tick();
				StatInc(theXactErrCount);
				break;
			}
		}
	}
	checkProgress();
}

bool Client::launchXact(const Address &addr) {
	ConnInfo *ci = connMgr()->get(addr);
	if (!ci)
		return false;
	Xaction *x = genXaction(*ci);
	if (!x)
		return false;
	startKid(x);
	return true;
}

Xaction *Client::genXaction(ConnInfo &ci) {
	Xaction *x = new ClXaction();
	x->configure(theCfg.theXactCfg.clone());
	x->conn(
		ci.getMx(ConnInfo::dirIn),
		ci.getMx(ConnInfo::dirOut));
	return x;
}
