// $Id: PhaseMgr.cc,v 1.1.2.8 1998/12/02 07:43:08 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"

#include "Clock.h"
#include "PhaseMgr.h"
#include "LocalStats.h"
#include "Log.h"
#include "globals.h"


/* Phase */

Phase::Phase(const char *aName): theName(0), theStats(0), theGoal(-1) {
	theStats = new LocalStats();
	name(aName);
}

Phase::~Phase() {
	delete theStats;
}

bool Phase::enough() const {
	return theGoal > 0 && complCount() >= theGoal;
}

void Phase::incGoal(int delta) {
	int g = theGoal + delta;
	while (g + delta <= complCount())
		g += delta;
	goal(g);
}

void Phase::start(LocalStats *&stats) {
	theStartTime = TheClock;
	theStopTime = Time();
	stats = theStats;
}

void Phase::stop() {
	theStopTime = TheClock;
}

void Phase::reset() {
	delete theStats;
	theStats = new LocalStats();
	theStartTime = theStopTime = Time();
	theGoal = -1;
}

void Phase::report(ostream &os, bool full) const {
	const Time len = theStopTime - theStartTime;
	/*SysLog << "#" << theName 
		<< " phase done in " << len
		<< " after " << TheComplCount
		<< " compl ( " << theStats->theComplMeter << " compl/sec )"
		<< endl;
	*/

	// use actual phase start/stop times, not first/last subm/compl time
	os << theName
		<< "\t " << theStopTime.sec()
		<< "\t " << setw(6) << TheComplCount
		<< "\t " << setw(6) << submCount()
		<< "\t " << setw(6) << Rate(submCount(), len)
		<< "\t " << setw(6) << complCount()
		<< "\t " << setw(6) << Rate(complCount(), len)
		<< "\t " << setw(6) << theStats->dhr()
		<< "\t " << setw(6) << theStats->theXactErrCount
		<< endl;

	if (full) {
		theStats->report(os << endl);
		reportRespTimeAndSize(os);
	}
}

void Phase::reportRespTimeAndSize(ostream &os) const {

	if (!theStats->theRespTimeTrace || !theStats->theRespSizeTrace) {
		os << "# response time and size stats are disabled" << endl;
		return;
	}

	theStats->theRespTimeTrace->report(os << endl);
	theStats->theRespSizeTrace->report(os << endl);

	// report response time versus size
	Array<int> perc;
	theStats->theRespSizeTrace->perc(perc); // get percentiles
	if (!perc.count()) {
		os << "# response time and size stats are not available" << endl;
		return;
	}
	Assert(perc.count() == 101);

	// print header
	os << endl << "Resp Time for major Resp Size ranges: {" << endl
		<< setw(6) << "perc"
		<< "\t " << setw(6) << "min_sz"
		<< "\t " << setw(6) << "max_sz";
	for (int j = 0; j <= 100; j += 25)
		os << "\t " << setw(5) << j << '%';
	os << endl;

	// XXX: a very inefficient and ugly loop
	for (int p = 10; p <= 100; p += 10) {
		const int min_size = perc[p-10];
		const int max_size = perc[p] - (int)(p < 100);

		// collect matching response times
		const int kept_count = theStats->theRespTimeTrace->keptCount();
		Assert(kept_count == theStats->theRespSizeTrace->keptCount());
		Trace T("tmp", "usec", kept_count);
		for (int i = 0; i < kept_count; ++i) {
			const int sz = theStats->theRespSizeTrace->item(i);
			if (sz >= min_size && sz <= max_size)
				T.record(theStats->theRespTimeTrace->item(i));
		}

		// print percentiles for this size range
		if (T.stat().count()) {
			Array<int> tp;
			T.perc(tp);
			if (!tp.count())
				continue;
			Assert(tp.count() == 101);
			os << setw(6) << p
				<< "\t " << setw(6) << min_size 
				<< "\t " << setw(6) << max_size;
			for (int j = 0; j <= 100; j += 25)
				os << "\t " << setw(6) << tp[j];
			os << endl;
		}
	}
	os << '}' << endl;
}

/* PhaseMgr */

PhaseMgr::PhaseMgr(): thePhase(0), thePinger(0), theState(rsNone) {
}

PhaseMgr::~PhaseMgr() {
	delete thePhase;
	delete thePinger;
}

// set goals (expected number of completions)
void PhaseMgr::start() {
	assert(theState == rsNone);

	thePhase = new Phase();
	thePinger = new Phase();

	thePinger->goal(1000);

	if (TheWorld.goal() > 0) {
		thePhase->name("warm-up");
		thePinger->name("1K-warm");
		thePhase->goal(TheWorld.goal()/10);
		theState = rsWarmUp;
	} else {
		thePhase->name("measurement");
		thePinger->name("1K-meas");
		thePhase->goal(-1);
		theState = rsMeasure;
	}

	thePhase->start(ThePhaseStats);
	thePinger->start(ThePingStats);
}

// checks metering phase and updates the PhaseMgr as needed
void PhaseMgr::checkPhase() {
	if (thePhase->enough())
		nextPhase();

	if (thePinger->enough()) {
		thePinger->stop();
		thePinger->report(Report);
		thePinger->reset();
		thePinger->goal(1000);
		thePinger->start(ThePingStats);
	}
}

void PhaseMgr::lastReport() {

	if (!thePhase->stopped())
		thePhase->stop();
	thePhase->report(Report,  theState <= rsMeasure);
}

void PhaseMgr::nextPhase() {
	assert(thePhase);

	// report results for the current phase
	thePhase->stop();
	thePhase->report(Report, theState == rsMeasure);

	// advance to the next phase
	Assert(theState < rsCoolOff);

	delete thePhase;
	theState = (RunState) (1 + (int)theState);

	if (theState == rsMeasure) {
		thePinger->name("1K-meas");
		thePhase = new Phase("measurement");
		thePhase->goal(TheWorld.goal() - TheWorld.goal()/10);
		thePhase->start(ThePhaseStats);
		ThePhaseStats->collectTraces();
	} else {
		thePinger->name("1K-cool");
		thePhase = new Phase("cool-off");
		thePhase->goal(-1);
		thePhase->start(ThePhaseStats);
	}
}

