#ifndef _IO_BUFFER_H_
#define _IO_BUFFER_H_ "$Id: IOBuffer.h,v 1.11.2.2 1998/12/15 07:28:05 rousskov Exp $"

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "Size.h"

#include <string.h>

// handy in-memory buffer that accumulates and "leaks" data as needed
class IOBuffer {
	public:
		typedef int State; // an opaque class with "==" and "!=" operations
		typedef IOBuffer &BufManip(IOBuffer &);

	public:
		IOBuffer();
		~IOBuffer();

		void reset(); // resets offsets and clears flags

		bool empty() const { return theSize <= theOffset; }

		Size contSize() const { return theSize - theOffset; }
		Size bufSize() const { return theCapacity - theSize; }
		Size inOffset() const { return theInOffset; }
		Size outOffset() const { return theOutOffset; }

		// Danger: these two "pointers" are for atomic operations only!!
		const char *content(int off = 0) const { return theBuf + theOffset + off; }
		char *buf() const { return theBuf + theSize; }

		// fiddle with buffer sizes
		void pack(); // moves content to the beginning of the buffer
		void stretch(Size sizeToAppend); // makes bufSize() >= sizeToAppend
		void minCapacity(Size minCap) { if (theCapacity < minCap) grow(minCap); }

		// call these after the "pointers" above where used
		void appended(Size size) { theSize += size; theInOffset += size; nextState(); if (theSize > theCapacity) abort(); }
		void consumed(Size size) { theOffset += size; theOutOffset += size; nextState(); if (theOffset >= theSize) clean(); }
		void consumedUpTo(const char *ptr) { consumed(ptr-content()); }

		void append(const char *str, Size size);
		void append(const char *str) { append(str, strlen(str)); }

		// convenience operators
		IOBuffer &operator <<(char c) { append(&c, 1); return *this; }
		IOBuffer &operator <<(const char *s) { append(s); return *this; }
		IOBuffer &operator <<(int n);

		IOBuffer &operator >>(char &c);
		IOBuffer &operator >>(char c);
		IOBuffer &operator >>(char *s);
		IOBuffer &operator >>(const char *s);
		IOBuffer &operator >>(int &n);

		void skipWhiteSpace();

		// to allow for external buf-manipulators
		IOBuffer& operator<<(BufManip func) { return (*func)(*this); }
		IOBuffer& operator>>(BufManip func) { return (*func)(*this); }

		// diagnostics
		bool exhausted() const { return err() || (eod() && empty()); }
		bool good() const { return !err(); }

		// state manipulation
		State state() const { return theState; }
		bool errFmt() const { return theFmtBug; }
		bool err() const { return theError != 0; }
		bool eod() const { return theEOD; }

		void errFmt(bool aFmtBug) { theFmtBug = aFmtBug; nextState(); }
		void err(int aError) { theError = aError; nextState(); }
		void eod(bool aEOD) { theEOD = aEOD; nextState(); }

		int errNo() const { return theError; }

	protected:
		void clean();
		void grow(Size neededCapacity);

		void nextState() { theState++; }

	protected:
		char *theBuf;

		Size theCapacity;  // total allocated memory
		Size theSize;      // the end of content (used memory)
		Size theOffset;    // the beginning of content
		Size theInOffset;  // cumulative size of all append()s since init
		Size theOutOffset; // cumulative size of all consume()s since init

		State theState;    // to detect changes in buffer "state"

		int theError;      // errno
		bool theFmtBug;    // parsing bugs
		bool theEOD;       // end-of-data
};

// manipulators
inline IOBuffer &ws(IOBuffer &buf) { buf.skipWhiteSpace(); return buf; }
inline IOBuffer &ends(IOBuffer &buf) { return buf << ((char)0); }
inline IOBuffer &endl(IOBuffer &buf) { return buf << '\n'; }
inline IOBuffer &crlf(IOBuffer &buf) { return buf << '\r' << '\n'; }

#endif
