// $Id: Rnd.h,v 1.6.2.2 1998/11/08 04:56:46 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#ifndef _RND_H_
#define _RND_H_

#include "AppDef.h"

#include <stdlib.h>
#include <math.h>

class ostream;

// holds current state of a random generator
typedef unsigned long RndState;

// a wrapper around random() generator
// every generator gets its own seed and state
class RndGen {
	public:
		RndGen(unsigned long seed = 1) { theState = seed; }
		~RndGen() {}

		// uniform[0,1)
		double trial() { return ltrial()/2147483648.0; }
		double operator ()() { return trial(); }

		// equiprob[lo,hi); note that "hi" is excluded
		double operator ()(double lo, double hi) { return lo + (hi-lo)*trial(); }
		int operator ()(int lo, int hi) { return (int)(lo + (hi-lo)*trial()); }

		// uniform[0,2147483648)
		long ltrial();

		RndGen *clone() const { return new RndGen(); }

		// change seed
		void seed(unsigned long seed);

	public://protected:
		RndState theState;
};

// abstract distribution
// allows for sharing of one random generator among several distributions
class RndDistr
{
	public:
		RndDistr(RndGen *aGen): theGen(aGen) {}
		virtual ~RndDistr() {} // XXX: leaking; no delete theGen;

		virtual const char *pdfName() const = 0; // type of pdf (e.g., exp)
		RndGen &rndGen() { return *theGen; }

		// returns next pseudo random number
		double operator ()() { return trial(); }
		virtual double trial() { return theGen->trial(); }

		virtual double mean() const = 0; // mean
		virtual double var() const = 0;  // variance
		double dev() const { return sqrt(var()); }  // standard deviation

		virtual RndDistr *clone() = 0;

	protected:
		RndGen *theGen;
};

// prints short distribution profile
extern ostream &operator <<(ostream &os, const RndDistr &rd);

// uniform[lo,hi) distribution
class UnifDistr: public RndDistr
{
	public:
		UnifDistr(RndGen *aGen, double aLo, double aHi): RndDistr(aGen), theLo(aLo), theHi(aHi) { Assert(theHi >= theLo); }

		virtual const char *pdfName() const { return "unif"; }

		virtual double mean() const { return (theLo+theHi)/2; }
		virtual double var() const { return range()*range()/12; }
		double range() const { return theHi-theLo; }

		virtual RndDistr *clone() { return new UnifDistr(theGen->clone(), theLo, theHi); }

		virtual double trial() { return theLo + range()*RndDistr::trial(); }

	protected:
		double theLo;
		double theHi;
};

// exponential(mean) distribution
class ExpDistr: public RndDistr
{
	public:
		ExpDistr(RndGen *aGen, double aMean): RndDistr(aGen), theMean(aMean) {}

		virtual const char *pdfName() const { return "exp"; }

		virtual double mean() const { return theMean; }
		virtual double var() const { return theMean*theMean; }

		virtual RndDistr *clone() { return new ExpDistr(theGen->clone(), theMean); }

		virtual double trial() { return -theMean * log(RndDistr::trial()); }

	protected:
		double theMean;
};

// normal(mean, var) distribution
class NormDistr: public RndDistr
{
	public:
		NormDistr(RndGen *aGen, double aMean, double aVar): RndDistr(aGen), theMean(aMean), theVar(aVar) {}

		virtual const char *pdfName() const { return "norm"; }

		virtual double mean() const { return theMean; }
		virtual double var() const { return theVar; }

		virtual RndDistr *clone() { return new NormDistr(theGen->clone(), theMean, theVar); }

		virtual double trial();

	protected:
		double theMean;
		double theVar;
};

// lognormal(mu, sigmaSq) distribution
class LognDistr: public NormDistr
{
	public:
		// creates using mean/var
		static LognDistr *ViaMean(RndGen *aGen, double aMean, double aVar);
		// creates using mu/sigma
		static LognDistr *ViaMu(RndGen *aGen, double aMu, double aSigmaSq) { return new LognDistr(aGen, aMu, aSigmaSq); }

	public:
		virtual const char *pdfName() const { return "logn"; }

		virtual double mean() const;
		virtual double var() const;

		virtual RndDistr *clone() { return ViaMu(theGen->clone(), NormDistr::mean(), NormDistr::var()); }

		virtual double trial() { return ::exp(NormDistr::trial()); }

	protected:
		LognDistr(RndGen *generator, double aMu, double aSigmaSq): NormDistr(generator, aMu, aSigmaSq) {}
};

#endif
